"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createMiddleware = void 0;
const jsonwebtoken_1 = require("jsonwebtoken");
const p_pipe_1 = __importDefault(require("p-pipe"));
const jwks_rsa_1 = __importDefault(require("jwks-rsa"));
const ms_1 = __importDefault(require("ms"));
const util_1 = require("./util");
const callback_middleware_1 = require("./callback-middleware");
const cookies_middleware_1 = require("./cookies-middleware");
const process_callback_middleware_1 = require("./process-callback-middleware");
const proxy_middleware_1 = require("./proxy-middleware");
const signin_middleware_1 = require("./signin-middleware");
const signout_middleware_1 = require("./signout-middleware");
const userinfo_middleware_1 = require("./userinfo-middleware");
const createMiddleware = (issuerMetadata, client, options, sessionStore) => {
    const { clientServerOptions, proxyOptions } = options;
    const { signInPath, callbackPath, userInfoPath, signOutPath, processCallbackPath } = clientServerOptions;
    if (!issuerMetadata.jwks_uri && !options.clientServerOptions.enableOauth2) {
        throw new Error('issuer must support jwks_uri');
    }
    const userInfoFromJwtService = (0, util_1.createUserInfoFromJwtService)((0, jwks_rsa_1.default)({
        cache: true,
        cacheMaxEntries: 5,
        cacheMaxAge: (0, ms_1.default)('8hr'),
        jwksUri: String(issuerMetadata.jwks_uri)
    }), (token, secretOrPublicKey, callback) => {
        (0, jsonwebtoken_1.verify)(token, secretOrPublicKey, callback);
    });
    const coreMiddlewares = [
        (0, util_1.withBreaker)((0, cookies_middleware_1.cookiesMiddleware)(options, sessionStore), options, 'cookieMiddleware'),
        (0, util_1.withBreaker)((0, signin_middleware_1.signInMiddleware)(client, signInPath, options, sessionStore), options, 'signInMiddleware'),
        (0, util_1.withBreaker)((0, callback_middleware_1.callbackMiddleware)(client, callbackPath, processCallbackPath, options), options, 'callbackMiddleware'),
        (0, util_1.withBreaker)((0, process_callback_middleware_1.processCallbackMiddleware)(client, processCallbackPath, options, sessionStore), options, 'processCallbackMiddleware'),
        (0, util_1.withBreaker)((0, signout_middleware_1.signOutMiddleware)(signOutPath, options, sessionStore), options, 'signOutMiddleware'),
        (0, util_1.withBreaker)((0, userinfo_middleware_1.userInfoMiddleware)({
            client,
            pathname: userInfoPath,
            sessionStore,
            options,
            userInfoFromJwtService
        }), options, 'userInfoMiddleware')
    ];
    const corePipeline = (0, p_pipe_1.default)(...coreMiddlewares);
    const proxyMiddlewares = [];
    if (proxyOptions.proxyPaths.length === proxyOptions.proxyHosts.length) {
        let excludeCookie = false;
        let excludeOriginHeaders = false;
        let useIdToken = false;
        const includeCookieFlag = proxyOptions.excludeCookie
            ? proxyOptions.proxyPaths.length ===
                proxyOptions.excludeCookie.length
            : false;
        const includeOriginHeadersFlag = proxyOptions.excludeOriginHeaders
            ? proxyOptions.proxyPaths.length ===
                proxyOptions.excludeOriginHeaders.length
            : false;
        const includeUseIdTokenFlag = proxyOptions.excludeCookie
            ? proxyOptions.proxyPaths.length ===
                proxyOptions.excludeCookie.length
            : false;
        for (let i = 0; i < proxyOptions.proxyPaths.length; i++) {
            const proxyHost = proxyOptions.proxyHosts[i];
            const proxyPath = proxyOptions.proxyPaths[i];
            if (includeCookieFlag) {
                excludeCookie = proxyOptions.excludeCookie[i];
            }
            if (includeOriginHeadersFlag) {
                excludeOriginHeaders = proxyOptions.excludeOriginHeaders[i];
            }
            if (includeUseIdTokenFlag) {
                useIdToken = proxyOptions.useIdToken[i];
            }
            proxyMiddlewares.push((0, util_1.withBreaker)((0, proxy_middleware_1.proxyMiddleware)({
                host: proxyHost,
                pathname: proxyPath,
                excludeCookie,
                excludeOriginHeaders,
                useIdToken,
                sessionStore,
                client
            }), options, `proxyMiddleware ${proxyPath}:${proxyHost}`));
        }
    }
    if (proxyMiddlewares.length > 0) {
        const proxyPipeline = (0, p_pipe_1.default)(...proxyMiddlewares);
        return (0, p_pipe_1.default)(corePipeline, proxyPipeline);
    }
    return corePipeline;
};
exports.createMiddleware = createMiddleware;
//# sourceMappingURL=index.js.map