"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.securePathCheckMiddleware = void 0;
const status_1 = require("../status");
const util_1 = require("./util");
const securePathCheckMiddleware = (options, sessionStore, client) => {
    return async (ctx) => {
        const parsedUrl = ctx.url;
        const { securedPaths, signInPath, errorPagePath } = options.clientServerOptions;
        if (securedPaths) {
            if (!ctx.sessionId) {
                ctx.log.debug('secure path check detected no session id');
                (0, util_1.redirectResponse)(`${errorPagePath}?${(0, status_1.qsOfErrorResponse)(status_1.AccessDeniedErrorResponse)}`, ctx.res);
                return false;
            }
            ctx.log.debug(`secure page check processing configured secured paths ${securedPaths.join(',')}`);
            const session = await sessionStore.get(ctx.sessionId);
            if (!session) {
                ctx.log.debug('secure path check detected no session in store');
                (0, util_1.redirectResponse)(signInPath, ctx.res);
                return false;
            }
            if (session.tokenSet && !session.tokenSet.expired()) {
                ctx.log.debug('secure path check detected an existing valid session');
                return true;
            }
            if (session.tokenSet && session.tokenSet.expired()) {
                ctx.log.debug('secure path check detected an existing expired session');
                if (session.tokenSet.refresh_token) {
                    ctx.tokenSet = await client.refresh(session.tokenSet);
                    await sessionStore.set(ctx.sessionId, {
                        tokenSet: ctx.tokenSet
                    });
                    ctx.log.debug('secure path refreshed token');
                    return true;
                }
            }
            for (const securedPath of securedPaths) {
                // if unauthenticated request is accessing a secure path
                if (String(parsedUrl.pathname)
                    .toLowerCase()
                    .startsWith(securedPath.toLowerCase())) {
                    ctx.log.debug(`Request to access secure path ${securedPath} detected. Redirecting to signin path`);
                    // eslint-disable-next-line no-await-in-loop
                    await sessionStore.set(ctx.sessionId, {
                        securedPathFromUrl: `${String(parsedUrl.pathname)}${parsedUrl.search && parsedUrl.search.trim() !== ''
                            ? parsedUrl.search
                            : ''}`
                    });
                    (0, util_1.redirectResponse)(signInPath, ctx.res);
                    return false;
                }
            }
        }
        else {
            ctx.log.trace('secure paths no configured. skipping secure path check');
        }
        return true;
    };
};
exports.securePathCheckMiddleware = securePathCheckMiddleware;
//# sourceMappingURL=secure-path-check-middleware.js.map