"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.executeRequest = exports.pathFromReferer = exports.withBreaker = exports.parseBody = exports.createUserInfoFromJwtService = exports.pathsMatch = exports.showResponse = exports.sendResponse = exports.sendJsonResponse = exports.redirectResponse = exports.clone = void 0;
const url_1 = require("url");
const node_fetch_1 = __importDefault(require("node-fetch"));
const status_1 = require("../status");
const clone = (json) => JSON.parse(JSON.stringify(json));
exports.clone = clone;
const { status_code: defaultErrorResponseStatusCode } = status_1.DefaultErrorResponse;
const redirectResponse = (location, res) => {
    res.writeHead(302, {
        Location: location
    });
    res.end();
};
exports.redirectResponse = redirectResponse;
const sendJsonResponse = (status, json, res) => {
    const payload = JSON.stringify(json);
    (0, exports.sendResponse)(status, {
        'Content-Type': 'application/json; charset=utf-8',
        'Content-Length': Buffer.byteLength(payload)
    }, payload, res);
};
exports.sendJsonResponse = sendJsonResponse;
const sendResponse = (status, headers, payload, res) => {
    if (headers.raw) {
        headers = (0, exports.clone)(headers.raw());
    }
    res.writeHead(status, headers);
    res.end(payload);
};
exports.sendResponse = sendResponse;
const showResponse = (status, payload, res) => {
    res.statusCode = status;
    res.end(payload);
};
exports.showResponse = showResponse;
const pathsMatch = (url, pathname) => {
    if (url.pathname) {
        return url.pathname
            .toLowerCase()
            .startsWith(pathname.trim().toLowerCase());
    }
    return false;
};
exports.pathsMatch = pathsMatch;
const createUserInfoFromJwtService = (jwksClientInstance, jwtVerify) => ({
    async userInfoFromJwt(token) {
        return new Promise((resolve, reject) => {
            function getKey(header, callback) {
                if (!header.kid || !header.alg) {
                    reject(new Error('invalid jwt header'));
                    return;
                }
                jwksClientInstance.getSigningKey(header.kid, (error, key) => {
                    const _key = key;
                    const signingKey = (_key === null || _key === void 0 ? void 0 : _key.getPublicKey()) ||
                        _key.rsaPublicKey;
                    callback(error, signingKey);
                });
            }
            jwtVerify(token, getKey, 
            // @ts-ignore
            (error, decoded) => {
                if (error) {
                    reject(error);
                    return;
                }
                resolve(decoded);
            });
        });
    }
});
exports.createUserInfoFromJwtService = createUserInfoFromJwtService;
const parseBody = async (request) => {
    return new Promise(resolve => {
        let body = '';
        request.on('data', (chunk) => {
            body += chunk;
        });
        request.on('end', () => {
            resolve(body);
        });
    });
};
exports.parseBody = parseBody;
const withBreaker = (mw, options, name) => {
    return async (ctx) => {
        if (ctx.url.pathname) {
            const { errorPagePath } = options.clientServerOptions;
            if ((0, exports.pathsMatch)(ctx.url, errorPagePath)) {
                ctx.log.trace(`skipping ${name} with path ${ctx.url.pathname} on error page match`);
                return ctx;
            }
            if (ctx.done) {
                ctx.log.trace(`skipping ${name} with path ${ctx.url.pathname} on done ${String(ctx.done)}`);
                return ctx;
            }
            ctx.log.trace(`invoking ${name} with path ${ctx.url.pathname}`);
            return mw(ctx);
        }
        return ctx;
    };
};
exports.withBreaker = withBreaker;
const pathFromReferer = (referer) => {
    const refererUrl = new url_1.URL(referer);
    return `${refererUrl.pathname}${refererUrl.search ? `${refererUrl.search}` : ''}`;
};
exports.pathFromReferer = pathFromReferer;
const executeRequest = async (executeParameters) => {
    const { token, excludeCookie, excludeOriginHeaders, host, pathname, ctx, method } = executeParameters;
    const { headers } = ctx.req;
    const { path } = ctx.url;
    let body = '';
    const normalizedMethod = method.trim().toLowerCase();
    const includeBody = normalizedMethod === 'post' ||
        normalizedMethod === 'put' ||
        normalizedMethod === 'patch' ||
        normalizedMethod === 'delete';
    if (includeBody) {
        body = await (0, exports.parseBody)(ctx.req);
    }
    let requestHeaders = (0, exports.clone)(headers);
    if (excludeOriginHeaders) {
        requestHeaders = {};
    }
    requestHeaders.authorization = `Bearer ${token}`;
    if (excludeCookie) {
        delete requestHeaders.cookie;
    }
    const requestPath = String(path).replace(pathname, '');
    const requestUrl = new url_1.URL(`${host}${requestPath}`);
    requestHeaders.host = requestUrl.hostname;
    ctx.log.debug(`fetching ${requestUrl.href}`);
    try {
        const response = await (0, node_fetch_1.default)(requestUrl.href, {
            method: normalizedMethod,
            headers: requestHeaders,
            body: includeBody ? body : undefined
        });
        (0, exports.sendResponse)(response.status, response.headers, await response.text(), ctx.res);
    }
    catch (error) {
        ctx.log.error('Error sending proxy request', error);
        (0, exports.sendJsonResponse)(defaultErrorResponseStatusCode, status_1.DefaultErrorResponse, ctx.res);
    }
};
exports.executeRequest = executeRequest;
//# sourceMappingURL=util.js.map