"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.createRequestListener = exports.RequestListenerEventEmitter = exports.RequestListenerEvents = void 0;
const url_1 = require("url");
const events_1 = require("events");
const openid_client_1 = require("openid-client");
const pino_1 = __importDefault(require("pino"));
const context_1 = require("./context");
const middleware_1 = require("./middleware");
const secure_path_check_middleware_1 = require("./middleware/secure-path-check-middleware");
exports.RequestListenerEvents = {
    STARTED: 'started',
    COMPLETED: 'completed',
    ERROR: 'error'
};
exports.RequestListenerEventEmitter = new events_1.EventEmitter();
const createRequestListener = async (options, sessionStore, requestHandler) => {
    const { clientServerOptions: { discoveryEndpoint, emitEvents }, clientMetadata } = options;
    let issuer;
    const emitEvent = (eventName, action) => {
        if (emitEvents) {
            exports.RequestListenerEventEmitter.emit(eventName);
        }
        if (action) {
            action();
        }
    };
    if (options.clientServerOptions.enableOauth2) {
        issuer = new openid_client_1.Issuer({
            issuer: 'oauth2',
            authorization_endpoint: options.clientServerOptions.authorizationEndpoint,
            token_endpoint: options.clientServerOptions.tokenEndpoint,
            userinfo_endpoint: options.clientServerOptions.userInfoEndpoint
        });
    }
    else {
        issuer = await openid_client_1.Issuer.discover(discoveryEndpoint);
    }
    const client = new issuer.Client(clientMetadata);
    const { loggerOptions, loggerDestination } = options;
    let log;
    log = loggerDestination
        ? (0, pino_1.default)(loggerOptions, loggerDestination)
        : (0, pino_1.default)(loggerOptions);
    const middleware = (0, middleware_1.createMiddleware)(issuer.metadata, client, options, sessionStore);
    const securePathCheck = (0, secure_path_check_middleware_1.securePathCheckMiddleware)(options, sessionStore, client);
    log.debug(`PKCE ${options.clientServerOptions.enablePKCE ? 'enabled' : 'disabled'}`);
    log.debug(`Oauth2 ${options.clientServerOptions.enableOauth2 ? 'enabled' : 'disabled'}`);
    /* eslint-disable promise/prefer-await-to-then */
    return (request, res) => {
        emitEvent(exports.RequestListenerEvents.STARTED);
        if (request.url) {
            const parsedUrl = (0, url_1.parse)(request.url, true);
            middleware((0, context_1.createContext)(request, res, parsedUrl, log))
                .then(async (ctx) => {
                if (ctx.done) {
                    log.info(`Context indicates 'done'. Skipping requestHandler for pathname ${String(parsedUrl.pathname)}`);
                    emitEvent(exports.RequestListenerEvents.COMPLETED);
                }
                else if (requestHandler) {
                    try {
                        const passToHandler = await securePathCheck(ctx);
                        if (passToHandler) {
                            log.debug(`invoking requestHandler with parsedQuery pathname ${String(parsedUrl.pathname)} for session ${String(ctx.sessionId)}`);
                            await requestHandler(request, res, parsedUrl);
                            emitEvent(exports.RequestListenerEvents.COMPLETED);
                        }
                        else {
                            log.debug('secure path check indicated to not pass request to handler');
                            emitEvent(exports.RequestListenerEvents.COMPLETED);
                        }
                    }
                    catch (error) {
                        log.error(error);
                        emitEvent(exports.RequestListenerEvents.ERROR);
                    }
                }
                else {
                    log.debug("Context indicates 'not done' and no request handler was provided");
                    emitEvent(exports.RequestListenerEvents.COMPLETED);
                }
            })
                .catch(error => {
                emitEvent(exports.RequestListenerEvents.ERROR, () => {
                    log.error(error);
                });
            });
        }
        else if (requestHandler) {
            log.debug('req url not defined. invoking requestHandler');
            requestHandler(request, res)
                .then(() => {
                log.debug('requestHandler complete');
            })
                .then(() => {
                emitEvent(exports.RequestListenerEvents.COMPLETED);
            })
                .catch(error => {
                emitEvent(exports.RequestListenerEvents.ERROR, () => {
                    log.error(error);
                });
            });
        }
        else {
            log.debug('req url & requestHandler not defined');
            emitEvent(exports.RequestListenerEvents.COMPLETED);
        }
    };
    /* eslint-enable promise/prefer-await-to-then */
};
exports.createRequestListener = createRequestListener;
//# sourceMappingURL=request-listener.js.map